"""
import `orchestrator` module content
"""
from .block import AsyncBlock, SyncBlock
from .flow import AsyncFlow, FlowBlock, FlowBuilder, SyncFlow
from .orchestrator import AsyncOrchestrator, SyncOrchestrator



"""
orch_serv - the library for creating services and
an orchestrator that organized their interaction
"""
from .msg import BaseOrchServMsg
from .orchestrator import (
    AsyncBlock,
    AsyncFlow,
    AsyncOrchestrator,
    FlowBlock,
    FlowBuilder,
    SyncBlock,
    SyncFlow,
    SyncOrchestrator,
)
from .service import (
    AsyncCommandHandlerPostProcessStrategy,
    AsyncCommandHandlerProcessStrategy,
    AsyncService,
    CommandHandlerPostProcessStrategy,
    CommandHandlerProcessStrategy,
    Service,
    ServiceBlock,
    ServiceBuilder,
)
from .stepper import Step, Stepper, StepsBuilder

__version__ = "0.1.4"
__all__ = [
    "__version__",
    "BaseOrchServMsg",
    "SyncBlock",
    "AsyncBlock",
    "SyncFlow",
    "AsyncFlow",
    "AsyncOrchestrator",
    "SyncOrchestrator",
    "FlowBlock",
    "FlowBuilder",
    "AsyncCommandHandlerPostProcessStrategy",
    "AsyncCommandHandlerProcessStrategy",
    "AsyncService",
    "CommandHandlerPostProcessStrategy",
    "CommandHandlerProcessStrategy",
    "Service",
    "ServiceBlock",
    "ServiceBuilder",
    "Stepper",
    "StepsBuilder",
    "Step",
]



"""
Module Service with help classes for build service
"""
from __future__ import annotations

from abc import ABC, abstractmethod
from logging import Logger
from typing import Any, Dict, List, Optional, Tuple, Type, Union

from pydantic import BaseModel

from orch_serv.exc import (
    DoublePostProcessFunctionDeclaredError,
    EmptyCommandsException,
    IncorrectDefaultCommand,
    NotUniqueCommandError,
    ServiceBlockException,
    ServiceBuilderException,
)
from orch_serv.msg import BaseOrchServMsg
from orch_serv.settings import DEFAULT_LOGGER


class CommandHandler(ABC):
    """
    Class with method for all handlers
    """

    _logger: Logger = DEFAULT_LOGGER  # for one global handler
    _service_instance: Service = None  # single scope for service_commands
    _is_logged = False

    @property
    def logger(self) -> Optional[Logger]:
        """
        get logger
        :return:  logger or None
        """
        return self._logger

    def set_logger(self, val: Any) -> None:
        """
        Set logger to handler
        :param Any val: logger
        :return: nothing
        :raise TypeError: if val is not a logger
        """
        if not val:
            return
        if isinstance(val, Logger):
            self._logger = val
        else:
            raise TypeError(f"Type must be logger but not {val}")

    def set_service_instance(self, service: Service) -> None:
        """
        set to instance handler current service object for use swap
        :param Service service:
        :return: nothing
        """
        if isinstance(service, Service):
            self._service_instance = service
        else:
            raise TypeError(f"Type must be Service but not {service}")

    def set_to_swap_scope(self, key: str, data: Any) -> bool:  # pragma: no cover
        """
        Method adds a value to the global scope for access from all services
        :param str key: name key
        :param Any data: data to add
        :return: bool: is_added_value
        """
        is_added = False
        if self._service_instance:
            try:
                setattr(self._service_instance, key, data)
                is_added = True
            except Exception as exc:
                if self.logger and self._is_logged:
                    self.logger.warning(
                        "Key `%s` not added. Error %s", key, str(exc), exc_info=True
                    )
        else:
            if self.logger and self._is_logged:
                self.logger.warning("You can't use swap because it is not initialized ")
        return is_added

    def get_from_swap_scope(self, key: str) -> Optional[Any]:  # pragma: no cover
        """
        Method gets a value available for all services from the global scope.
        :param str key: name key to get value
        :return: data if exist or None
        """
        data = None
        if self._service_instance:
            try:
                data = getattr(self._service_instance, key)
            except Exception as exc:
                if self.logger and self._is_logged:
                    self.logger.warning(
                        "Error while fetching data from swap by key %s. Error %s",
                        key,
                        str(exc),
                        exc_info=True,
                    )
        else:
            if self.logger and self._is_logged:
                self.logger.warning(
                    "You cann't use swap because it is not initialized "
                )
        return data

    def del_from_swap_scope(self, key: str) -> bool:  # pragma: no cover
        """
        Method removes attribute from swap_scope if any
        """
        is_dropped = False
        if self._service_instance:
            try:
                delattr(self._service_instance, key)
                is_dropped = True
            except Exception as exc:
                if self.logger and self._is_logged:
                    self.logger.warning(
                        "Key `%s` not deleted. Error %s", key, str(exc), exc_info=True
                    )
        else:
            if self.logger and self._is_logged:
                self.logger.warning("You can't use swap because it is not initialized ")
        return is_dropped


class CommandHandlerProcessStrategy(CommandHandler, ABC):
    """
    Handler class for base processing message
    """

    @property
    def target_command(self):  # pragma: no cover
        """
        this command will determine that the message should be processed
        by this particular service
        must be unique within the service
        """
        raise NotImplementedError

    @abstractmethod
    def process(
        self, msg: BaseOrchServMsg
    ) -> Union[Tuple[BaseOrchServMsg, Any], BaseOrchServMsg]:  # pragma: no cover
        """
        the main method for executing the logic of this handler, must be overridden
        in the inheritor
        :param MessageQueue msg: msg from queue
        :return: MessageQueue or None if return None post handler will not be called
        """
        raise NotImplementedError


class CommandHandlerPostProcessStrategy(CommandHandler, ABC):
    """
    Post Process Handler
    """

    @abstractmethod
    def post_process(
        self, msg: BaseOrchServMsg, additional_data: Optional[Any] = None
    ) -> None:  # pragma: no cover
        """
        method for post-processing
        e.g. sending to another queue
        , must be overridden in the inheritor
        :param MessageQueue msg:
        :param additional_data: optional data got from processing block
        :return: None
        """
        raise NotImplementedError


class AsyncCommandHandlerProcessStrategy(CommandHandler, ABC):
    """
    Handler class for base processing message
    """

    @property
    def target_command(self):  # pragma: no cover
        """
        this command will determine that the message
        should be processed by this particular service
        must be unique within the service
        """
        raise NotImplementedError

    @abstractmethod
    async def process(
        self, msg: BaseOrchServMsg
    ) -> Union[Tuple[BaseOrchServMsg, Any], BaseOrchServMsg]:  # pragma: no cover
        """
        the main method for this handler execution logic, must be overridden
        in the inheritor
        :param MessageQueue msg: msg from queue
        :return: MessageQueue or None if return None post handler will not be called
        """
        raise NotImplementedError


class AsyncCommandHandlerPostProcessStrategy(CommandHandler, ABC):
    """
    Post Process Handler
    """

    @abstractmethod
    async def post_process(
        self, msg: BaseOrchServMsg, additional_data: Optional[Any] = None
    ) -> None:  # pragma: no cover
        """
        method does post-processing
        e.g. sending to another queue
        , must be overridden in the inheritor
        :param MessageQueue msg:
        :param additional_data: optional data got from processing block
        :return: None
        """
        raise NotImplementedError


class DefaultPostProcessStrategy(CommandHandlerPostProcessStrategy):
    def post_process(self, msg: BaseOrchServMsg, additional_data: Optional[Any] = None):
        """
        default post process function for sync service
        :param BaseOrchServMsg msg: msg received after processing
        :param Optional[Any] additional_data:
        :return: None
        """
        pass


class AsyncDefaultPostProcessStrategy(AsyncCommandHandlerPostProcessStrategy):
    async def post_process(
        self, msg: BaseOrchServMsg, additional_data: Optional[Any] = None
    ) -> None:
        """
        default post process function for async service
        :param BaseOrchServMsg msg: msg received after processing
        :param Optional[Any] additional_data:
        :return: None
        """
        pass


class ServiceCommand(BaseModel):
    """
    Structure class
    """

    processor: Union[CommandHandlerProcessStrategy, AsyncCommandHandlerProcessStrategy]
    post_processor: Union[
        CommandHandlerPostProcessStrategy, AsyncCommandHandlerPostProcessStrategy
    ]

    class Config:
        arbitrary_types_allowed = True


class ServiceBlock:
    """
    Class for build service - one Block with process
     and post process handlers
    """

    _processor: Union[
        CommandHandlerProcessStrategy, AsyncCommandHandlerProcessStrategy
    ] = None
    _post_processor: Optional[
        Union[CommandHandlerPostProcessStrategy, AsyncCommandHandlerPostProcessStrategy]
    ] = None

    @property
    def processor(
        self,
    ) -> Union[CommandHandlerProcessStrategy, AsyncCommandHandlerProcessStrategy]:
        """
        Property contains process handler
        :return: process command instance
        :rtype: Union[CommandHandlerProcessStrategy, AsyncCommandHandlerProcessStrategy]
        """
        return self._processor

    @property
    def post_processor(
        self,
    ) -> Union[
        CommandHandlerPostProcessStrategy, AsyncCommandHandlerPostProcessStrategy
    ]:
        """
        Property contains post_process handler
        :return: post_processor instance
        :rtype:Union[CommandHandlerPostProcessStrategy,
         AsyncCommandHandlerPostProcessStrategy]
        """
        return self._post_processor

    @processor.setter  # type: ignore # noqa
    def processor(  # noqa
        self,
        processor: Union[
            CommandHandlerProcessStrategy,
            AsyncCommandHandlerProcessStrategy,
            Type[CommandHandlerProcessStrategy],
            Type[AsyncCommandHandlerProcessStrategy],
        ],
    ) -> None:
        """
        Setups service processor and checks type
        :param processor:
        :raise ServiceBlockException:
        :return:
        """
        exc = ServiceBlockException(
            f"`process` object must be of type "
            f"`CommandHandlerProcessStrategy` or `AsyncCommandHandlerProcessStrategy` "
            f"and not {type(processor)}"
        )
        if isinstance(processor, type):
            if issubclass(
                processor,
                (CommandHandlerProcessStrategy, AsyncCommandHandlerProcessStrategy),
            ):
                self._processor = processor()
            else:
                raise exc
        elif isinstance(
            processor,
            (CommandHandlerProcessStrategy, AsyncCommandHandlerProcessStrategy),
        ):
            self._processor = processor
        else:
            raise exc

    @post_processor.setter  # type: ignore # noqa
    def post_processor(  # noqa
        self,
        post_processor: Optional[
            Union[
                CommandHandlerPostProcessStrategy,
                AsyncCommandHandlerPostProcessStrategy,
                Type[CommandHandlerPostProcessStrategy],
                Type[AsyncCommandHandlerPostProcessStrategy],
            ]
        ],
    ) -> None:
        """
        Setups service processor and checks type
        :param post_processor:
        :raise ServiceBlockException:
        :return:
        """
        if post_processor is None:
            return
        exc = ServiceBlockException(
            f"`post_processor` object must be of type "
            f"`CommandHandlerPostProcessStrategy` "
            f"or `AsyncCommandHandlerPostProcessStrategy`"
            f" and not {type(post_processor)}"
        )
        if isinstance(post_processor, type):
            if issubclass(
                post_processor,
                (
                    CommandHandlerPostProcessStrategy,
                    AsyncCommandHandlerPostProcessStrategy,
                ),
            ):
                self._post_processor = post_processor()
            else:
                raise exc
        elif isinstance(
            post_processor,
            (CommandHandlerPostProcessStrategy, AsyncCommandHandlerPostProcessStrategy),
        ):
            self._post_processor = post_processor
        else:
            raise exc

    def __init__(
        self,
        processor: Union[
            Type[CommandHandlerProcessStrategy],
            CommandHandlerProcessStrategy,
            Type[AsyncCommandHandlerProcessStrategy],
            AsyncCommandHandlerProcessStrategy,
        ],
        post_processor: Union[
            Type[CommandHandlerPostProcessStrategy],
            CommandHandlerPostProcessStrategy,
            Type[AsyncCommandHandlerPostProcessStrategy],
            AsyncCommandHandlerPostProcessStrategy,
        ] = None,
    ):
        """
        Init ServiceBlock
        :param CommandHandlerProcessStrategy process:
        :param CommandHandlerPostProcessStrategy post_process:
        """
        self.processor = processor  # type: ignore # noqa
        self.post_processor = post_processor  # type: ignore # noqa


class ServiceBuilder:
    """
    Class for aggregating service commands
    """

    _default_post_processor: Union[
        CommandHandlerPostProcessStrategy, AsyncCommandHandlerPostProcessStrategy
    ] = None

    @property
    def default_post_processor(
        self,
    ) -> Union[
        CommandHandlerPostProcessStrategy, AsyncCommandHandlerPostProcessStrategy
    ]:
        return self._default_post_processor

    @default_post_processor.setter
    def default_post_processor(
        self,
        default_post_processor: Union[
            CommandHandlerPostProcessStrategy,
            Type[CommandHandlerPostProcessStrategy],
            AsyncCommandHandlerPostProcessStrategy,
            Type[AsyncCommandHandlerPostProcessStrategy],
        ],
    ) -> None:
        """
        Setups service processor and checks correct type
        :param post_processor:
        :raise ServiceBlockException: if incorrect type
        :return:
        """
        if default_post_processor:
            message_error = (
                f"`default_post_processor` object must be of type "
                f"`CommandHandlerPostProcessStrategy` "
                f"and not {type(default_post_processor)}"
            )
            self._default_post_processor = self.check_is_post_processor(
                default_post_processor, message_error
            )

    @staticmethod
    def check_is_post_processor(obj: Any, message_error: str):
        """
        Method checks if obj is post_processor
        :param obj:
        :param message_error:
        :raise ServiceBlockException: if incorrect type

        :return:
        """
        exc = ServiceBlockException(message_error)
        if isinstance(obj, type):
            if issubclass(
                obj,
                (
                    CommandHandlerPostProcessStrategy,
                    AsyncCommandHandlerPostProcessStrategy,
                ),
            ):
                return obj()
            else:
                raise exc
        elif isinstance(
            obj,
            (CommandHandlerPostProcessStrategy, AsyncCommandHandlerPostProcessStrategy),
        ):
            return obj
        else:
            raise exc

    def __init__(
        self,
        *args: ServiceBlock,
        default_post_process: Optional[
            Union[
                Type[CommandHandlerPostProcessStrategy],
                CommandHandlerPostProcessStrategy,
                Type[AsyncCommandHandlerPostProcessStrategy],
                AsyncCommandHandlerPostProcessStrategy,
            ]
        ] = None,
    ):
        self.default_post_processor = default_post_process  # type: ignore # noqa
        list_blocks = list()  # type: List[ServiceBlock]
        for service_block in args:
            if isinstance(service_block, ServiceBlock):
                list_blocks.append(service_block)
            else:
                try:
                    post_processor = self.check_is_post_processor(service_block, "")
                    if self.default_post_processor:
                        raise DoublePostProcessFunctionDeclaredError()
                    else:
                        self.default_post_processor = post_processor
                except ServiceBlockException:
                    raise TypeError(
                        f"block must be instance class `ServiceBlock`."
                        f" Not {type(service_block)}"
                    )
        if not list_blocks:
            raise EmptyCommandsException()
        self._list_blocks = list_blocks

    def build(
        self, service_instance: Service, logger: Optional[Logger] = None
    ) -> Dict[str, ServiceCommand]:
        """
        Method builds all commands' current service.
        :param service_instance: current service
        :param Optional[Logger] logger: logger
        :return: {command_name: ServiceCommand} dict commands and proccess classes
        :rtype: Dict[str, ServiceCommand]
        """
        logger = logger or DEFAULT_LOGGER
        dict_commands = dict()  # type: Dict[str, ServiceCommand]

        if not self.default_post_processor:
            if service_instance._base_process_class == CommandHandlerProcessStrategy:
                self.default_post_processor = DefaultPostProcessStrategy()
            else:
                self.default_post_processor = AsyncDefaultPostProcessStrategy()

        self.default_post_processor.set_logger(logger)
        self.default_post_processor.set_service_instance(service_instance)

        for block in self._list_blocks:
            processor = block.processor
            post_processor = block.post_processor
            processor.set_logger(logger)  # noqa
            processor.set_service_instance(service_instance)
            if post_processor:
                post_processor.set_logger(logger)  # noqa
                post_processor.set_service_instance(service_instance)

            else:
                post_processor = self.default_post_processor

            if dict_commands.get(processor.target_command):
                raise NotUniqueCommandError(
                    f"Command `{processor.target_command}`"
                    f" is not unique for current service"
                )
            else:

                dict_commands[processor.target_command] = ServiceCommand(
                    processor=processor, post_processor=post_processor
                )
        return dict_commands


class Service(ABC):
    """
    Class Service for handle msg-s
    """

    _service_commands: ServiceBuilder = None
    _dict_handlers: Dict[str, ServiceCommand] = dict()
    _default_command: str = None
    _base_process_class = CommandHandlerProcessStrategy
    _base_post_process_class = CommandHandlerPostProcessStrategy

    def __init__(
        self,
        service_commands: Optional[ServiceBuilder] = None,
        default_command: Optional[str] = None,
        logger: Optional[Logger] = None,
        is_catch_exceptions: bool = True,
    ):
        self._is_catch_exceptions = is_catch_exceptions
        self.logger = logger or DEFAULT_LOGGER
        if not service_commands:
            # for validate if use property setup
            service_commands = self.service_commands

        self.__validate_service_builder(service_commands)
        self._dict_handlers = self.service_commands.build(
            service_instance=self, logger=logger
        )
        if default_command:
            if not self._dict_handlers.get(default_command):
                raise IncorrectDefaultCommand(
                    self._default_command, list(self._dict_handlers.keys())
                )
            else:
                self._default_command = default_command
        self._validate_data()

    def __validate_service_builder(self, service_builder: ServiceBuilder):
        """
        Help function for service_builder object validation
        :param service_builder:
        :return:
        """
        if not isinstance(service_builder, ServiceBuilder):
            raise ServiceBuilderException(
                f"Variable `service_builder` must be a"
                f" ServiceBuilder and not {type(service_builder)}"
            )
        self._service_commands = service_builder

    @property
    def service_commands(self) -> ServiceBuilder:  # pragma: no cover
        if not self._service_commands:
            raise NotImplementedError
        return self._service_commands

    @service_commands.setter
    def service_commands(self, service_builder: ServiceBuilder) -> None:
        self.__validate_service_builder(service_builder)

    def _validate_data(self):  # pragma: no cover
        """
        Help function to validate data service
        """
        if not issubclass(self._base_process_class, CommandHandler):
            raise TypeError(
                f"Incorrect type `_base_processor_class`. "
                f"Parent of variable `_base_processor_class` must be `CommandHandler`"
                f" and not {self._base_process_class.__base__}. "
                f"Please don`t override protected variable."
            )
        if not issubclass(self._base_post_process_class, CommandHandler):
            raise TypeError(
                f"Incorrect type `_base_post_processor_class`."
                f" Parent of variable `_base_post_processor_class` "
                f"must be `CommandHandler`"
                f" and not {self._base_post_process_class.__base__}. "
                f"Please don`t override protected variable."
            )
        for command_name, command in self._dict_handlers.items():
            if not isinstance(command.processor, self._base_process_class):
                raise TypeError(
                    f"Incorrect type `processor` of command `{command_name}`."
                    f" Must be a `{self._base_process_class.__name__}` "
                    f"and not `{command.processor.__class__.__base__.__name__}`"
                )
            if not isinstance(command.post_processor, self._base_post_process_class):
                raise TypeError(
                    f"Incorrect type `post_processor` of command `{command_name}`. "
                    f"Must be a `{self._base_post_process_class.__name__}` "
                    f"and not `{command.post_processor.__class__.__base__.__name__}`"
                )

    def _get_service_command(
        self, message: BaseOrchServMsg
    ) -> Optional[ServiceCommand]:
        """
        Function for get command to handle received message
        :param BaseOrchServMsg message: received message
        :return: handler if exist
        """
        command = self._dict_handlers.get(message.get_command())
        if command:
            return command
        elif self._default_command:
            return self._dict_handlers[self._default_command]
        else:
            return None

    def handle(
        self, message: BaseOrchServMsg, is_force_return: bool = False
    ) -> Optional[BaseOrchServMsg]:
        """
        the main function to process received messages
        :param BaseOrchServMsg message: message to process
        :param bool is_force_return: return msgs all time after execution
        :return: not processed msgs or received msgs if is_force_return==True
        """
        is_return_message = is_force_return
        self.logger.info("Service. Started processing message %s", message)
        command = self._get_service_command(message)
        if command:
            try:
                resp_process = command.processor.process(message)
                if resp_process:
                    try:
                        resp_msg, additional_data = resp_process  # type: ignore # noqa
                    except TypeError:
                        resp_msg, additional_data = resp_process, None
                    if command.post_processor and resp_msg:
                        command.post_processor.post_process(resp_msg, additional_data)
                else:
                    self.logger.debug(
                        "Don't send to post-processing because"
                        " processor doesn't return data."
                    )
            except Exception as exc:
                is_return_message = True
                self.logger.warning(
                    "Error in time processing msg %s. Error %s",
                    str(message),
                    str(exc),
                    exc_info=True,
                )
                if not self._is_catch_exceptions:
                    raise exc
        else:
            is_return_message = True
            self.logger.warning("Not found command to process message %s", str(message))

        self.logger.info("Service. Finished processing message %s", message)
        if is_return_message:
            return message
        return None


class AsyncService(Service, ABC):
    """ """

    _base_process_class = AsyncCommandHandlerProcessStrategy  # type: ignore # noqa
    _base_post_process_class = AsyncCommandHandlerPostProcessStrategy  # type: ignore # noqa

    async def handle(  # type: ignore # noqa
        self, message: BaseOrchServMsg, is_force_return: bool = False
    ) -> Optional[BaseOrchServMsg]:
        """

        :param message:
        :param is_force_return:
        :return:
        """
        is_return_message = is_force_return
        self.logger.info("Service. Started processing message %s", message)
        command = self._get_service_command(message)
        if command:
            try:
                resp_process = await command.processor.process(message)  # type: ignore # noqa
                if resp_process:
                    try:
                        resp_msg, additional_data = resp_process
                    except TypeError:
                        resp_msg, additional_data = resp_process, None
                    if command.post_processor and resp_msg:
                        await command.post_processor.post_process(
                            resp_msg, additional_data
                        )
                else:
                    self.logger.debug(
                        "Don't send to post-processing because "
                        "processor don't return data."
                    )
            except Exception as exc:
                is_return_message = True
                self.logger.warning(
                    "Error in time processing msg %s. Error %s",
                    str(message),
                    str(exc),
                    exc_info=True,
                )
                if not self._is_catch_exceptions:
                    raise exc
        else:
            is_return_message = True
            self.logger.warning("Not found command to process message %s", str(message))

        self.logger.info("Service. Finished processing message %s", message)
        if is_return_message:
            return message
        return None



"""
Module with async orchestrator
"""
from typing import Optional

from orch_serv.msg import BaseOrchServMsg
from orch_serv.orchestrator.block import AsyncBlock
from orch_serv.orchestrator.flow import AsyncFlow

from .sync_orchestrator import SyncOrchestrator


class AsyncOrchestrator(SyncOrchestrator):
    """
    AsyncOrchestrator
    override handle function for async mode
    """

    _base_class_for_flow = AsyncFlow  # noqa
    _base_class_for_target = AsyncBlock  # noqa

    async def handle(  # type: ignore
        self, message: BaseOrchServMsg, is_force_return: bool = False
    ) -> Optional[BaseOrchServMsg]:
        """
        Message processing method

        !!! the presence of a `target` in the message has higher priority
         than the presence of a `flow`
        :param message: message to process
        :type message: BaseOrchServMsg
        :param is_force_return: always return a message
        :type is_force_return: bool

        :return: message if an error occurred while processing the message
         if is_force_return is False
        :rtype: Optional[BaseOrchServMsg]
        """
        is_return_message = is_force_return
        self.logger.debug("Orchestrator. Started processing msg: %s", str(message))
        if not isinstance(message, BaseOrchServMsg):
            raise TypeError(
                "Incorrect type `message`."
                " The `message` the message must be "
                f"of type `BaseOrchServMsg` and not {type(message)}"
            )
        if message.get_flow() or message.get_target():
            if message.get_target():
                name_target = message.get_target()
                target = self._targets.get(name_target)
                if not target and not self._default_block:
                    is_return_message = True
                    self.logger.warning(
                        "Orchestrator. No suitable target was found to process "
                        "this message with target: %s."
                        " Available targets: %s",
                        name_target,
                        list(self._targets.keys()),
                    )
                else:
                    if not target:
                        target = self._targets.get(self._default_block)
                    if isinstance(target, type):
                        target = target(logger=self.logger)
                        self._targets[target.name_block] = target
                    try:
                        await target.process(message)
                    except Exception as exc:
                        is_return_message = True
                        self.logger.warning(
                            "Orchestrator. Error processing msg `%s` in target `%s`."
                            " Error: %s",
                            str(message),
                            name_target,
                            str(exc),
                            exc_info=True,
                        )
            else:
                name_flow = message.get_flow()
                flow = self._flows.get(name_flow)
                if not flow and not self._default_flow:
                    is_return_message = True
                    self.logger.warning(
                        "Orchestrator. No suitable flow was found to process "
                        "this message with flow: %s."
                        " Available flows: %s",
                        name_flow,
                        list(self._flows.keys()),
                    )
                else:
                    if not flow:
                        flow = self._flows.get(self._default_flow)
                    if isinstance(flow, type):
                        flow = flow(logger=self.logger)
                        self._flows[flow.name_flow] = flow
                    try:
                        await flow.to_go_with_the_flow(message)
                    except Exception as exc:
                        is_return_message = True
                        self.logger.warning(
                            "Orchestrator. Error processing msg %s in flow %s."
                            " Error: %s",
                            str(message),
                            name_flow,
                            str(exc),
                            exc_info=True,
                        )
        else:
            is_return_message = True
            self.logger.warning(
                "Orchestrator. Msg %s. "
                "Does not contain information about the type of"
                " processing and it was not processed",
                str(message),
            )
        self.logger.debug("Orchestrator. Finished processing msg: %s", str(message))
        if is_return_message:
            return message
        return None



"""
Module with block class from which flow chains are formed
"""
from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Awaitable, Callable, Optional

from orch_serv.msg import BaseOrchServMsg


class SyncBaseBlock(ABC):
    """
    Base class for handling sync blocks
    """

    @property
    def pre_handler_function(
        self,
    ) -> Optional[
        Callable[[BaseOrchServMsg], Optional[BaseOrchServMsg]]
    ]:  # pragma: no cover
        """
        The function that will be executed before the main handler
         must return a message if the message is not returned,
         then the message will not get into the handler
        """
        raise NotImplementedError

    @property
    def post_handler_function(
        self,
    ) -> Optional[
        Callable[[BaseOrchServMsg], Optional[BaseOrchServMsg]]
    ]:  # pragma: no cover
        """
        function to be executed after the main handler
        """
        raise NotImplementedError

    @property
    def name_block(self):  # pragma: no cover
        """
        Unique name to identify block
        for override in subclass   name_block
        """
        raise NotImplementedError

    @abstractmethod
    def set_next(self, handler: SyncBaseBlock) -> SyncBaseBlock:  # pragma: no cover
        """
        Save Next handler after this handler in flow
        :param  handler: block for execution after current
        :type  handler: Union[AsyncBaseBlock, Type[AsyncBaseBlock]]
        :return: AsyncBaseBlock
        :raise Exception: some exception if error is in time init handler if
         handler provided as type
        :raise TypeError: if handler is not instance of type SyncBaseBlock
        """
        raise NotImplementedError

    @abstractmethod
    def get_next(self) -> SyncBaseBlock:  # pragma: no cover
        """
        the method returns the next block after
        :return: next block if exist
        :rtype: Optional[AsyncBaseBlock]
        """
        raise NotImplementedError

    @abstractmethod
    def get_list_flow(self) -> str:  # pragma: no cover
        """
        Method return str steps flow
        :return: str
        """
        raise NotImplementedError

    @abstractmethod
    def handle(self, message: BaseOrchServMsg) -> None:  # pragma: no cover
        """
        A function that determines which handler should now process
         the message based on the source from which the message came.
        :param message: message for processing
        :type message: BaseOrchServMsg
        :return: nothing
        :raise FlowException: exception if a message with the wrong source
         is passed to the flow
        """
        raise NotImplementedError

    @abstractmethod
    def process(
        self, message: BaseOrchServMsg
    ) -> Optional[BaseOrchServMsg]:  # pragma: no cover
        """
        Function to be redefined in subclasses which contains
         the main logic of this block
        If necessary, after executing this function, execute
         the post function method that must return a message
        :param message: message to process
        :type message: BaseOrchServMsg
        :return: message after processing
        :rtype: Optional[BaseOrchServMsg]
        """
        raise NotImplementedError("Not Implemented method for processing messages")


class AsyncBaseBlock(ABC):
    """
    Base class for handling async blocks
    """

    @property
    def pre_handler_function(
        self,
    ) -> Optional[
        Callable[[BaseOrchServMsg], Awaitable[Optional[BaseOrchServMsg]]]
    ]:  # pragma: no cover
        """
        The function that will be executed before the main handler.
         Must return a message, if the message is not returned,
         then the message will not get into the handler
        """
        raise NotImplementedError

    @property
    def post_handler_function(
        self,
    ) -> Optional[
        Callable[[BaseOrchServMsg], Awaitable[Optional[BaseOrchServMsg]]]
    ]:  # pragma: no cover
        """
        function to be executed after the main handler
        """
        raise NotImplementedError

    @property
    def name_block(self):  # pragma: no cover
        """
        Unique name to identify block
        to override in subclass name_block
        """
        raise NotImplementedError

    @abstractmethod
    def set_next(self, handler: AsyncBaseBlock) -> AsyncBaseBlock:  # pragma: no cover
        """
        Save next handler after this handler in flow
        :param  handler: block for execution after current
        :type  handler: Union[AsyncBaseBlock, Type[AsyncBaseBlock]]
        :return: AsyncBaseBlock
        :raise Exception: some exception if error in time init handler if
         handler provided as type
        :raise TypeError: if handler not is instance of type SyncBaseBlock
        """
        raise NotImplementedError

    @abstractmethod
    def get_next(self) -> Optional[AsyncBaseBlock]:  # pragma: no cover
        """
        the method returns the next block after
        :return: next block if exist
        :rtype: Optional[AsyncBaseBlock]
        """

        raise NotImplementedError

    @abstractmethod
    def get_list_flow(self) -> str:  # pragma: no cover
        """
        Method return str steps flow
        :return: str
        """
        raise NotImplementedError

    @abstractmethod
    async def handle(self, message: BaseOrchServMsg) -> None:  # pragma: no cover
        """
        A function that determines which handler should now process
         the message based on the source from which the message came.
        :param message: message for processing
        :type message: BaseOrchServMsg
        :return: nothing
        :raise FlowException: exception if a message with the wrong source
         is passed to the flow
        """
        raise NotImplementedError

    @abstractmethod
    async def process(
        self, message: BaseOrchServMsg
    ) -> Optional[BaseOrchServMsg]:  # pragma: no cover
        """
        Function to be redefined in subclasses which contains
         the main logic of this block
        If necessary, after executing this function, execute
         the post function method must return a message
        :param message: message to process
        :type message: BaseOrchServMsg
        :return: message after processing
        :rtype: Optional[BaseOrchServMsg]
        """
        raise NotImplementedError("Not Implemented method for processing messages")



"""
Module consolidating all exceptions lib
"""

# pylint: disable=non-parent-init-called, super-init-not-called
from typing import Any, List, Optional


class OrchServError(Exception):
    """
    Main exception class for custom exception lib
    """


class MsgException(OrchServError):
    """
    Main exception class for exceptions with msg
    """


class OrchestratorException(OrchServError):
    """
    Main exception class exceptions with Orchestrator
    """


class FlowException(OrchestratorException):
    """
    Error if the handler for the current message
     is not found in the flow
    """

    def __init__(self, message: str):
        self.message = (
            f"The flow chain ended without finding a single handler: {message}"
        )
        Exception.__init__(self, self.message)


class FlowBlockException(OrchestratorException):
    """
    Class custom exception
    for incorrect type of flow block
    """

    def __init__(self, message: str = ""):
        self.message = (
            f"Incorrect type block handler. "
            f"The block must inherit from the class 'Block',"
            f" and not : {message}"
        )
        Exception.__init__(self, self.message)


class FlowBuilderException(OrchestratorException):
    """
    Class custom exception
    for wrong types
    """

    def __init__(self, message: str = ""):
        self.message = f"Incorrect type in builder arguments {message}"
        Exception.__init__(self, self.message)


class UniqueNameException(OrchestratorException):
    """
    Exception for not unique flows
    """

    def __init__(self, not_unique_flow_name: str, _type: str):
        self.message = (
            f"The {_type} name `{not_unique_flow_name}` "
            f"is not unique to this orchestrator_service"
        )
        Exception.__init__(self, self.message)


class NoDateException(OrchestratorException):
    """
    Exception if dict flow is empty
    """

    def __init__(self, _type, msg: Optional[str] = None):
        if msg:
            self.message = msg
        else:
            self.message = f"No {_type}s for processing"
        Exception.__init__(self, self.message)


class WrongTypeException(OrchestratorException):
    """
    Exception for incorrect input types
    """

    def __init__(self, variable: str = "flows", type_variable: str = "any"):
        self.message = (
            f"Invalid variable type `{variable}`."
            f" There {variable} should be a list and not `{type_variable}`"
        )
        Exception.__init__(self, self.message)


class WorkTypeMismatchException(OrchestratorException):
    """
    if an asynchronous class is used for
     a synchronous orchestra, or vice versa
    """

    def __init__(self, base_class: str, obj_class: str, is_target: bool = False):
        self.message = (
            f"In {'targets' if is_target else 'flows'} use incorrect mode class."
            f"For {base_class} use {obj_class}."
            f" Different types of work (sync and async or vice versa)"
        )

        Exception.__init__(self, self.message)


class NotFoundDefaultError(OrchestratorException):
    """
    If provided default value but default value does not exist in processed data
    """

    def __init__(
        self, default_value: str, allowed_values: List[str], is_target: bool = False
    ):
        self.message = (
            f"Not exist default value {default_value}"
            f" for {'targets' if is_target else 'flows'}."
            f"Allowed values: {allowed_values}"
        )
        Exception.__init__(self, self.message)


class NotUniqueBlockInFlowError(OrchestratorException):
    """
    If provided default value but default value does not exist in processed data
    """

    def __init__(self, block_name: str, flow_name: str):
        self.message = (
            f"Not unique block {block_name}"
            f" in flow {flow_name}."
            f"Check your flow or set flow "
            f"`variable is_contains_duplicat_blocks` as True"
        )
        Exception.__init__(self, self.message)


class ServiceException(OrchServError):
    """
    Main exception class exceptions with Service
    """


class ServiceBlockException(ServiceException):
    """
    Class for exceptions in ServiceBlock
    """


class DoublePostProcessFunctionDeclaredError(ServiceException):
    """
    Exception if many default postprocess handlers
    """

    def __init__(self):
        self.message = "Several postprocessors specified"
        Exception.__init__(self, self.message)


class EmptyCommandsException(ServiceException):
    """
    Empty list of commands for service operation
    """

    def __init__(
        self,
    ):
        self.message = "Empty list of commands for service operation"
        Exception.__init__(self, self.message)


class NotUniqueCommandError(ServiceException):
    """
    Class error if user add the same commands
    """

    pass


class ServiceBuilderException(ServiceException):
    """
    Class for exceptions in BuilderService
    """

    pass


class IncorrectDefaultCommand(ServiceException):
    """
    Among the available commands, there is no default command
    """

    def __init__(self, command: str, list_command: List[str]):
        self.message = (
            f"The `{command}` command which is the default "
            f"command is not among the valid commands : {str(list_command)}"
        )
        Exception.__init__(self, self.message)


class StepperException(OrchServError):
    """
    Main exception class exceptions with Stepper
    """


class NoDataForExecutionStepException(StepperException):
    """
    Exception for empty response
    """

    def __init__(
        self,
        step: str,
    ):
        self.message = (
            f"Object run step `{step}` didn't return data. "
            f"Check object logic or change settings Stepper.is_execute_if_empty = True."
        )

        Exception.__init__(self, self.message)


class ConsistencyStepsException(StepperException):
    """
    handling attribute error
    for cases when the data returned by the
     previous object + additional arguments declared during
     block initialization do not fit the method structure
    """

    def __init__(
        self,
        step: str,
        previous_step: str,
        signature_step_obj: str,
        return_annotation_previous_step_obj: str,
        received_from_previous_step: Any,
        args_on_init_step: Any,
    ):
        self.message = (
            f"Error occurred while trying to execute a step {step}. "
            f"The error is related to the attributes passed to the method."
            f"Method expects the following signature `({signature_step_obj})`. "
            f"From previous step `{previous_step}`->"
            f" {return_annotation_previous_step_obj}"
            f" received value(s) {received_from_previous_step} "
            f"additional arguments passed during step"
            f" initialization {args_on_init_step}. "
            f"Check the reality of the input and output data."
        )

        Exception.__init__(self, self.message)


class DataConsistencyError(StepperException):
    """
    Error when expected and passed data do not match
    """

    def __init__(self, obj: str, errors: List[str]):
        self.message = f"Error during check object{obj}. Error(s):\n {','.join(errors)}"
        Exception.__init__(self, self.message)


class ExtraAttributeError(StepperException):
    """
    If in kwargs for function provided extra attributes
    """

    def __init__(self, obj: str, extra_attributes: List[str], obj_attributes: str):
        self.message = (
            f"When declaring a step for an object `{obj}`"
            f" extra attribute(s) were passed `{','.join(extra_attributes)}`."
            f" Allowed attribute(s): {obj_attributes}"
        )
        Exception.__init__(self, self.message)



"""
Module with classes for build flow
"""
# pylint: disable=no-else-return,too-few-public-methods
from __future__ import annotations

import types
from logging import Logger
from typing import Callable, Optional, Type, Union

from orch_serv.exc import (
    FlowBlockException,
    FlowBuilderException,
    NotUniqueBlockInFlowError,
    WorkTypeMismatchException,
)
from orch_serv.orchestrator.block import AsyncBlock, SyncBlock
from orch_serv.orchestrator.block.base_block import AsyncBaseBlock, SyncBaseBlock


class FlowBlock:
    """
    Block for FlowBuilder
    """

    obj_block: Union[SyncBlock, AsyncBlock, type] = None

    def __init__(
        self,
        obj_block: Union[SyncBlock, AsyncBlock, Type[Union[SyncBlock, AsyncBlock]]],
        pre_handler_function: Optional[
            Union[str, types.FunctionType, types.MethodType, Callable]
        ] = None,
        post_handler_function: Optional[
            Union[str, types.FunctionType, types.MethodType, Callable]
        ] = None,
    ):
        """
        Init FlowBlock
        :param obj_block: block for flow
        :type obj_block: Union[SyncBlock, AsyncBlock,
         Type[Union[SyncBlock, AsyncBlock]]]
        :param pre_handler_function: optional function for execution before block
        :param  post_handler_function: optional function for execution after block
        """
        try:
            if isinstance(obj_block, type) and issubclass(
                obj_block, (SyncBlock, AsyncBlock)
            ):
                self.obj_block = obj_block
                self.pre_handler_function = pre_handler_function
                self.post_handler_function = post_handler_function
                return

            elif isinstance(obj_block, (SyncBlock, AsyncBlock)):
                self.obj_block = obj_block
                self.pre_handler_function = pre_handler_function
                self.post_handler_function = post_handler_function
                return

            raise FlowBlockException(str(type(obj_block)))
        except FlowBlockException as exc:
            raise exc
        except Exception:  # noqa
            raise TypeError("Incorrect type `obj_block`") from Exception

    @staticmethod
    def _get_function(
        instance_main: Flow,
        function_to_get: Optional[
            Union[str, types.FunctionType, types.MethodType, Callable]
        ],
    ) -> Optional[Union[types.FunctionType, types.MethodType, Callable]]:
        """
        helper function that returns a function if specified for the block
        :param instance_main:
        :type instance_main: Flow
        :param function_to_get: name of function or the function itself
        :type function_to_get: Union[str, types.FunctionType,
         types.MethodType, Callable]
        :return: a function object if provided
        :rtype: Optional[Union[types.FunctionType, types.MethodType, Callable]]
        """
        if isinstance(function_to_get, str):
            return getattr(instance_main, function_to_get, None)
        else:
            return function_to_get

    def init_block(
        self, instance_main: Flow, step_number: int = 0
    ) -> Union[SyncBlock, AsyncBlock]:
        """
        Method init instance subclass MainBlock
        :param instance_main: flow object for which this
         block is initialized
        :type instance_main: Flow
        :param int step_number: sequence number of the current block
        :return: object subclass MainBlock
        :rtype: Union[SyncBlock, AsyncBlock]
        """
        if not isinstance(instance_main, Flow):
            raise TypeError("Value `instance_main` must be a Flow")
        result: Union[AsyncBlock, SyncBlock]
        if isinstance(self.obj_block, type):

            result = self.obj_block(
                pre_handler_function=self._get_function(
                    instance_main, self.pre_handler_function
                ),
                post_handler_function=self._get_function(
                    instance_main, self.post_handler_function
                ),
                logger=instance_main.logger,
            )
            self.obj_block = result
        else:
            self.obj_block.pre_handler_function = self._get_function(  # type: ignore
                instance_main, self.pre_handler_function
            )
            self.obj_block.post_handler_function = self._get_function(  # type: ignore
                instance_main, self.post_handler_function
            )
            self.obj_block.logger = instance_main.logger
            result = self.obj_block
        if instance_main.is_contains_duplicat_blocks:
            result.name_block = result.name_block + f"_{step_number}"  # type: ignore
        return result


class FlowBuilder:
    """
    Flow building class
    build chain flow from its blocks
    """

    def __init__(self, step: FlowBlock, *args):
        """
        Init FlowBuilder
        :param FlowBlock step: first block in flow
        :param List[FlowBlock] args:  other steps  if value exist
        """
        self.steps = list(args)
        self.steps.insert(0, step)
        for _index, _step in enumerate(self.steps):
            if not isinstance(_step, FlowBlock):
                raise FlowBuilderException(f"on index {_index}")

    def build_flow(self, instance_main: Flow) -> Union[SyncBlock, AsyncBlock]:
        """
        Build chain flow for Flow
        :param instance_main: current flow
        :type instance_main: Flow
        :return: the first block in the flow chain
        :rtype: Union[SyncBlock, AsyncBlock]
        """

        flow = self.steps[0].init_block(instance_main, 0)
        cur_step = flow
        for i, step in enumerate(self.steps[1:]):
            cur_step = cur_step.set_next(step.init_block(instance_main, i + 1))
        return flow


class Flow:
    """
    Class for inheritance for a specific flow
    :attr flow_chain: flow chain starting from the first block
    :type flow_chain: Union[SyncBlock, AsyncBlock]
    :attr is_contains_duplicat_blocks: whether the flow contains repeating blocks
     needed to avoid looping
    :type is_contains_duplicat_blocks: bool

    """

    flow_chain: Optional[Union[SyncBaseBlock, AsyncBaseBlock]] = None
    is_contains_duplicat_blocks: bool = False

    @property
    def name_flow(self) -> str:  # pragma: no cover
        """
        Unique name to identify flow
        for override in subclass 'name_flow'
        :return: name flow
        """
        raise NotImplementedError

    @property
    def _base_class_for_blocks(
        self,
    ) -> Type[Union[AsyncBlock, SyncBlock]]:  # pragma: no cover
        """
        An additional property for child classes to make the flow work only
         with synchronous or asynchronous blocks.
         necessary to maintain integrity
        :return:
        """
        raise NotImplementedError

    @property
    def steps_flow(self):  # pragma: no cover
        """
        blocks that make up the current flow
        :return:
        """
        raise NotImplementedError

    @steps_flow.setter
    def steps_flow(self, flow: FlowBuilder):  # pragma: no cover
        """
        check the set value to property `steps_flow` value
        :param FlowBuilder flow: builder flow for current flow
        :return: None or exception
        """
        if isinstance(flow, FlowBuilder):
            self.steps_flow = flow
        else:
            raise TypeError("incorrect type flow builder")

    def __init__(self, logger: Optional[Logger] = None):
        """
        Init Flow
        :param logger: orchestrator logger
        """
        self.logger = logger or Logger(__name__)
        self.logger.info("Init block %s", self.name_flow)

        if isinstance(self.steps_flow, FlowBuilder):
            self.flow_chain = self.steps_flow.build_flow(self)
        else:
            raise TypeError(
                f"Incorrect type 'steps_flow' - it must be 'FlowBuilder',"
                f" and not {type(self.steps_flow)}"
            )
        self._validate_data()

    def _validate_data(self) -> None:
        """
        flow validation function after initialization
        """
        current = self.flow_chain
        list_exists_blocks = list()
        while current:
            if not isinstance(current, self._base_class_for_blocks):
                raise WorkTypeMismatchException(
                    base_class=self.__class__.__name__,
                    obj_class=current.__class__.__name__,
                    is_target=False,
                )
            if current.name_block in list_exists_blocks:
                raise NotUniqueBlockInFlowError(
                    block_name=current.name_block, flow_name=self.name_flow
                )
            list_exists_blocks.append(current.name_block)
            current = current.get_next()

    def get_steps(self) -> str:
        """
        Print steps flow
        :return:
        """
        return self.flow_chain.get_list_flow()



"""Module with base msg for processing"""
# pylint: disable=too-few-public-methods,no-name-in-module

from typing import Generic, Optional, TypeVar

from pydantic import BaseModel, Field
from pydantic.generics import GenericModel

SubPydanticBodyModel = TypeVar("SubPydanticBodyModel", bound=BaseModel)
SubPydanticHeaderModel = TypeVar("SubPydanticHeaderModel", bound=BaseModel)


class BaseOrchServMsg(
    GenericModel, Generic[SubPydanticBodyModel, SubPydanticHeaderModel]
):
    """
    Base message class for processing and to use the library
    :attr body: message body to be processed with basic information
    :type SubPydanticBodyModel:
    :attr header: message header with data to be processed
    :type SubPydanticHeaderModel:
    :example:
    >>> # How to use BaseOrchServMsg to create a custom message for your needs
    >>> # The first step: create body model your msg
    >>> from pydantic import BaseModel
    >>> class MyBody(BaseModel):
    >>>     body_option: str # any type you need
    >>> # The second step: create header model your
    >>> # msg (is optional if you need a header)
    >>> class MyHeader(BaseModel):
    >>>     header_option: str # any type you need
    >>> # The third step: create your msg
    >>> # The first option:
    >>> MyTypeFir = BaseOrchServMsg[MyBody, MyHeader]
    >>> example_my_type = {"body": {
    >>>                         "body_option": "test_body"},
    >>>                    "header":{
    >>>                          "header_option":"test_header"}}
    >>> my_value = MyTypeFir(**example_my_type)
    >>> # First option without header
    >>> MyTypeFirWithoutHeader = BaseOrchServMsg[MyBody, BaseModel]
    >>> # The second option
    >>> class MyTypeSec(BaseOrchServMsg):
    >>>     body: MyBody
    >>>     header: MyHeader
    >>> MyTypeSec(**example_my_type)
    """

    body: SubPydanticBodyModel = Field(
        ..., description="The body of the message with the " "structure you need"
    )
    header: Optional[SubPydanticHeaderModel] = Field(
        description="Optional message header " "with the structure you need"
    )

    def get_source(self) -> Optional[str]:
        """
        ### For orchestrator ###
        Adding the source to the message where
          the message was processed (will be processed)

        Method is required to be implemented
        pair method to method set_source

        :raise NotImplementedError: if not override
        :return: the method returns the string that the set method
        :rtype: str
        :example:
        >>> obj: BaseOrchServMsg
        >>> source = 'test_source'
        >>> obj.set_source(source)
        >>> assert obj.get_source() == source
        >>> # True
        """
        raise NotImplementedError

    def set_source(self, source: str) -> None:
        """
        ### For orchestrator ###
        Method set source to message to identify previous step on flow
        Method is required to be implemented
        for more information see method get_source

        :param source: source where the message was processed (will be processed)
        :type source: str
        :raise NotImplementedError: if not override method
        :return: nothing
        """
        raise NotImplementedError

    def get_flow(self) -> Optional[str]:
        """
        ### For orchestrator ###
        Method returns the name of the flow to which the given message belongs
        !!! the message must contain a field that determines whether
         the message belongs to some flow
        :return: title flow
        :rtype: str
        """
        raise NotImplementedError

    def get_target(self) -> Optional[str]:
        """
        ### For orchestrator ###
        Method returns the name of the command that should execute the given message
        mandatory if you don’t need a flow but want to execute only one
         block for a message
        :return: title block to execution
        :rtype: str
        """
        raise NotImplementedError

    def get_command(self) -> Optional[str]:
        """
        ### For service ###
        Method returns the name of the processor
        that should execute the given message
        :return: title processor to execution
        :rtype: str
        """
        raise NotImplementedError



"""
Flow for sync mode
"""
from abc import ABC

from orch_serv.msg import BaseOrchServMsg

from .. import AsyncBlock
from .base_flow import Flow


class AsyncFlow(Flow, ABC):
    """
    AsyncFlow for execution async blocks
    """

    _base_class_for_blocks = AsyncBlock
    flow_chain: AsyncBlock = None

    async def to_go_with_the_flow(self, message: BaseOrchServMsg) -> None:
        """
        Method that starts flow execution from the first block
        :param message: message to process
        :type message: BaseOrchServMsg
        :return: None
        """
        await self.flow_chain.handle(message)



"""
Module with base block class for user sync blocks
"""
# pylint: disable=not-callable
from __future__ import annotations

import types
from abc import ABC
from copy import deepcopy
from logging import Logger
from typing import Callable, Optional, Type, Union

from orch_serv.exc import FlowException
from orch_serv.msg import BaseOrchServMsg

from .base_block import SyncBaseBlock


class SyncBlock(SyncBaseBlock, ABC):
    """
    Base block for sync mode
    :attr _next_handler: the object of the next handler
     in the flow where this block participates
    :type _next_handler: SyncBaseBlock
    :attr _pre_handler_function: function called before this block's handler
    :type _pre_handler_function:  Optional[
        Callable[[BaseOrchServMsg], Optional[BaseOrchServMsg]]
    ]
    :attr _post_handler_function: function called after this block's handler
    :type _post_handler_function:  Optional[
        Callable[[BaseOrchServMsg], Optional[BaseOrchServMsg]]
    ]
    """

    _next_handler: SyncBaseBlock = None
    _pre_handler_function: Optional[
        Callable[[BaseOrchServMsg], Optional[BaseOrchServMsg]]
    ] = None
    _post_handler_function: Optional[
        Callable[[BaseOrchServMsg], Optional[BaseOrchServMsg]]
    ] = None

    @property
    def is_execute_after_nullable_process_msg(self) -> bool:
        """
        Execute if the block handler did not return a message
        :return: true if it should be executed with previous msg after empty process msg
        :rtype: bool
        """
        return True

    @property
    def pre_handler_function(
        self,
    ) -> Optional[Callable[[BaseOrchServMsg], Optional[BaseOrchServMsg]]]:
        """
        A class property that returns an synchronous function that will be
         called before being sent to the main handler.
        the function should only take one parameter msg: BaseOrchServMsg
        Attention!!!
        The function must return a message object.
         If the function does not return a message, then the handler will not be called
        :return: sync function if exist pre_handler_function
        :rtype: Optional[
            Callable[[BaseOrchServMsg], Optional[BaseOrchServMsg]]]
        """
        return self._pre_handler_function

    @property
    def post_handler_function(
        self,
    ) -> Optional[Callable[[BaseOrchServMsg], Optional[BaseOrchServMsg]]]:
        """
        A class property that returns an synchronous function that will be
         called after this block`s handler.
        the function should only take one parameter msg: BaseOrchServMsg
        if the `process` does not return a message to the function,
         the message sent to the handler will be transferred
        if it is not mandatory to execute, redefine
         the variable `is_execute_after_nullable_process_msg = False` in your block
        :return: sync function if exist pre_handler_function
        :rtype: Optional[
         Callable[[BaseOrchServMsg], Optional[BaseOrchServMsg]]]
        """
        return self._post_handler_function

    @pre_handler_function.setter  # type: ignore
    def pre_handler_function(
        self,
        func: Optional[Callable[[BaseOrchServMsg], Optional[BaseOrchServMsg]]] = None,
    ):
        """
        Method check pre_handler_function is func
        :param func: object to check
        :type func: Callable
        :raise TypeError: if func is not method or function
        :return: nothing
        """
        if not func:
            pass
        elif isinstance(func, (types.FunctionType, types.MethodType)):
            self._pre_handler_function = func
        else:
            raise TypeError(
                "Incorrect type pre_handler_function,"
                " the attribute must be a function or None"
            )

    @post_handler_function.setter  # type: ignore # noqa
    def post_handler_function(
        self,
        func: Optional[Callable[[BaseOrchServMsg], Optional[BaseOrchServMsg]]] = None,
    ):
        """
        Method check post_handler_function is func
        :param func: object to check
        :type func: Callable
        :raise TypeError: if func is not method or function
        :return: nothing
        """
        if not func:
            pass
        elif isinstance(func, (types.FunctionType, types.MethodType)):
            self._post_handler_function = func
        else:
            raise TypeError(
                "Incorrect type post_handler_function,"
                " the attribute must be a function or None"
            )

    def __init__(
        self,
        pre_handler_function: Optional[
            Callable[[BaseOrchServMsg], Optional[BaseOrchServMsg]]
        ] = None,
        post_handler_function: Optional[
            Callable[[BaseOrchServMsg], Optional[BaseOrchServMsg]]
        ] = None,
        logger: Optional[Logger] = None,
    ):
        """
        Init Block
        :param pre_handler_function: function should accept
         and return objects of type Message
        which be run before call method
        :param post_handler_function: function should accept
         and return objects of type Message
        which be run after got msg from source
        :param logger: orchestrator logger
        """
        self.logger = logger or Logger(__name__)
        self.pre_handler_function = pre_handler_function  # type: ignore # noqa
        self.post_handler_function = post_handler_function  # type: ignore # noqa

    def set_next(
        self, handler: Union[SyncBaseBlock, Type[SyncBaseBlock]]
    ) -> SyncBaseBlock:
        """
        Save next handler after this handler in flow
        :param  handler: block for execution after current
        :type  handler: Union[SyncBaseBlock, Type[SyncBaseBlock]]
        :return: SyncBaseBlock
        :raise Exception: some exception if error in time init handler if
         handler provided as type
        :raise TypeError: if handler not is instance of type SyncBaseBlock
        """
        if isinstance(handler, type):
            try:
                handler = handler()
            except Exception as exc:
                raise exc
        if not isinstance(handler, SyncBaseBlock):
            raise TypeError("Incorrect type for next handler")

        self._next_handler = handler
        return handler

    def get_next(self) -> Optional[SyncBaseBlock]:
        """
        the method returns the next block
        :return: next block if exist
        :rtype: Optional[SyncBaseBlock]
        """
        return self._next_handler

    def _process_logic(self, block: SyncBaseBlock, message: BaseOrchServMsg) -> None:
        """
        Auxiliary function in which the logic of working with additional
         functions is hidden
        :param block: block for processing
        :type block: SyncBaseBlock
        :param message: message for processing
        :type message: BaseOrchServMsg
        :return: nothing
        """
        if block.pre_handler_function:
            message = block.pre_handler_function(message)
        if not message:
            return
        message.set_source(block.name_block)
        copy_msg = deepcopy(message)
        new_msg = block.process(message)
        if block.post_handler_function:
            if new_msg:
                block.post_handler_function(new_msg)
            elif self.is_execute_after_nullable_process_msg:
                block.post_handler_function(copy_msg)

    def handle(self, message: BaseOrchServMsg) -> None:
        """
        A function that determines which handler should now process
         the message based on the source from which the message came.
        :param message: message for processing
        :type message: BaseOrchServMsg
        :return: nothing
        :raise FlowException: exception if a message with the wrong source
         is passed to the flow
        """
        if not message.get_source():
            self._process_logic(block=self, message=message)
        elif message.get_source() == self.name_block:
            self._process_logic(block=self._next_handler, message=message)
        elif self._next_handler:
            self._next_handler.handle(message)
        else:
            raise FlowException(f"Not found block for source: {message}")

    def get_list_flow(self) -> str:
        """
        Method return str flow
        :return: str
        """
        if self._next_handler:
            next_blocks = self._next_handler.get_list_flow()
        else:
            next_blocks = "end"
        return f"{self.name_block} -> {next_blocks}"



"""
import `service` module content
"""
from .service import (
    AsyncCommandHandlerPostProcessStrategy,
    AsyncCommandHandlerProcessStrategy,
    AsyncService,
    CommandHandlerPostProcessStrategy,
    CommandHandlerProcessStrategy,
    Service,
    ServiceBlock,
    ServiceBuilder,
)



from __future__ import annotations

from copy import deepcopy
from logging import Logger
from typing import Any, Callable, Dict, List, Optional

from orch_serv.exc import ConsistencyStepsException, NoDataForExecutionStepException
from orch_serv.settings import DEFAULT_LOGGER
from orch_serv.stepper.utils import (
    get_returned_value,
    parse_signature,
    validate_data_consistency,
    validate_data_step,
)


class Step:
    """
    Class for configuration Step of Stepper flow
    :attr __obj: obj object that contains logic
    :type __obj: Callable
    :attr __kwargs: additional kwargs for execution step
     which are not returned from the previous step
    :type __kwargs: Dict[str, Any]
    :example:
    >>> def example_function(val: int , arg_1: int = 1) -> int:
    >>>     return val + arg_1
    >>> step_1 = Step(example_function )
    >>> step_1.execute(1) # return: 2
    >>> step_2 = Step(example_function, arg_1=2 )
    >>> step_1.execute(1) # return: 3

    """

    __obj: Callable = None
    __kwargs: Dict[str, Any] = dict()

    @property
    def obj(self) -> Callable:
        return self.__obj

    @property
    def kwargs(self) -> Dict[str, Any]:
        return self.__kwargs

    def __init__(self, obj: Callable, **kwargs):
        if not callable(obj):
            raise TypeError("Step `obj` must be a callable")
        self.__obj = obj  # type: ignore
        if kwargs:
            self.__kwargs = kwargs
        validate_data_step(self.__obj, self.__kwargs)

    def __repr__(self):  # pragma: no cover
        return str(self)

    def __str__(self):
        return f"<Step: obj: {self.obj.__name__}; kwargs:{self.__kwargs}>"

    def execute(self, data_to_execute: Any) -> Any:
        if not isinstance(data_to_execute, tuple):
            data_to_execute = (data_to_execute,)
        return self.obj(*data_to_execute, **self.kwargs)


class StepsBuilder:
    """
    Class for build and handling steps
    :attr __steps: steps flow
    :type __steps: List[Step]
    :example:
    >>> step_builder = StepsBuilder(
    >>>    Step(function1),
    >>>    Step(function2),
    >>>    is_validate_consistency_steps=True)
    """

    __steps: List[Step]

    def __init__(self, *steps: Step, is_validate_consistency_steps: bool = True):
        """
        Init StepBuilder
        :param steps: List of steps
        :type steps: *Step
        :param is_validate_consistency_steps: whether to check
         the consistency of steps by types
         and attributes or not
         !!! Attention !!!
         Used for methods with type annotation
         if you are not using type annotations then start or pass a value to False.
         !!! Then we are not responsible for the complete correctness of your data.
        :type is_validate_consistency_steps: bool default: True
        """
        self.__steps = list()
        if not isinstance(steps[0], Step):  # pragma: no cover
            raise TypeError(f"Step must be a Step and not {type(steps[0])}")
        self.__steps.append(steps[0])
        returned_previous_steps = get_returned_value(steps[0].obj)
        for step in steps[1:]:
            if not isinstance(step, Step):
                raise TypeError(f"Step must be a Step and not {type(step)}")
            if is_validate_consistency_steps:  # pragma: no cover
                validate_data_consistency(
                    obj=step.obj,
                    return_previous_obj=returned_previous_steps,
                    additional_args=deepcopy(step.kwargs),
                )
                returned_previous_steps = get_returned_value(step.obj)
            self.__steps.append(step)

    @property
    def steps(self) -> List[Step]:
        return self.__steps

    def __iter__(self):
        return StepsIterator(self)

    def __len__(self):
        return len(self.__steps)

    def __getitem__(self, index):
        return self.__steps[index]


class StepsIterator:
    def __init__(self, steps: StepsBuilder):
        if not isinstance(steps, StepsBuilder):
            raise TypeError
        self.__position = 0
        self.__steps = steps.steps

    def __next__(self) -> Step:
        if self.__position < len(self.__steps):
            value = self.__steps[self.__position]
            self.__position += 1
            return value
        else:
            raise StopIteration


class Stepper:
    """
    class for forming a flow where the input data
     of step n is the output of step n
    """

    __steps: StepsBuilder = None
    __is_execute_if_empty: bool = False

    @property
    def steps(self) -> StepsBuilder:  # pragma: no cover
        """
        Property contains a class with all blocks for the current flow
        :return: StepsBuilder
        """
        if self.__steps:  # pragma: no cover
            return self.__steps
        raise NotImplementedError

    @property
    def is_execute_if_empty(self) -> bool:
        """
        execute the next step with the data of the previous one
         if the current step did not return anything
        :return:
        """
        return self.__is_execute_if_empty

    def __init__(
        self,
        steps: Optional[StepsBuilder] = None,
        is_execute_if_empty: Optional[bool] = None,
        logger: Optional[Logger] = None,
    ):
        self.logger = logger or DEFAULT_LOGGER
        if steps:
            self.__steps = steps
        if not (is_execute_if_empty is None):
            self.__is_execute_if_empty = is_execute_if_empty
        if not isinstance(self.steps, StepsBuilder):
            raise TypeError(
                f"`Stepper.steps` must be a StepsBuilder and not {type(self.steps)} "
            )

    def __step_by_step(
        self,
        *args: Any,
    ) -> Any:
        step_data = args
        previous_step = "start"
        result_data: Any = None
        for i, step in enumerate(self.steps):
            self.logger.info("Starting Step %s", step)
            try:
                # if not isinstance(step_data, tuple):
                #     step_data = (step_data,)
                # result_data = step.obj(*step_data, **step.kwargs)
                result_data = step.execute(step_data)
                if not result_data and not self.is_execute_if_empty:
                    raise NoDataForExecutionStepException(
                        step.obj.__name__,
                    )
                elif not result_data:
                    continue
                else:
                    step_data = result_data
                    previous_step = step.obj.__name__

            except (AttributeError, TypeError) as exc:
                self.logger.error(
                    "Error in time processing %s. "
                    "Inconsistency of arguments for function "
                    "execution or internal object error."
                    " Error: %s",
                    step,
                    str(exc),
                    exc_info=True,
                )
                params_prev, return_prev = parse_signature(self.steps[i - 1].obj)
                params_current, return_current = parse_signature(self.steps[i].obj)
                raise ConsistencyStepsException(
                    step=self.steps[i].obj.__name__,
                    previous_step=previous_step,
                    signature_step_obj=params_current,
                    return_annotation_previous_step_obj=return_prev,
                    received_from_previous_step=step_data,
                    args_on_init_step=str(self.steps[i].kwargs),
                )
            except Exception as exc:
                self.logger.error(
                    "Error in time processing %s. Error: %s", step, str(exc)
                )
                raise exc

            self.logger.info("Finished Step %s", step)

        return result_data

    def step_by_step(
        self,
        *args: Any,
    ) -> Any:
        return self.__step_by_step(*args)


# ToDo add if else and loop !



"""
Module with sync orchestrator
"""
import inspect
from logging import Logger
from types import ModuleType
from typing import Any, Dict, List, Optional, Tuple, Type, Union

from orch_serv.exc import (
    NoDateException,
    NotFoundDefaultError,
    UniqueNameException,
    WorkTypeMismatchException,
    WrongTypeException,
)
from orch_serv.msg import BaseOrchServMsg
from orch_serv.orchestrator.block import AsyncBlock, SyncBlock
from orch_serv.orchestrator.flow import AsyncFlow, SyncFlow
from orch_serv.settings import DEFAULT_LOGGER


class SyncOrchestrator:
    """
    Orchestrator class for build service
    """

    _base_class_for_flow: Type[Union[SyncFlow, AsyncFlow]] = SyncFlow
    _base_class_for_target: Type[Union[SyncBlock, AsyncBlock]] = SyncBlock
    _default_flow: str = None
    _default_block: str = None

    _flows: Dict[
        str, Union[Type[Union[SyncFlow, AsyncFlow]], Union[SyncFlow, AsyncFlow]]
    ] = dict()
    _targets: Dict[
        str, Union[Type[Union[SyncBlock, AsyncBlock]], Union[SyncBlock, AsyncBlock]]
    ] = dict()

    @property
    def flows(self) -> Optional[Union[ModuleType, List]]:
        return None

    @property
    def blocks(self) -> Optional[Union[ModuleType, List]]:
        return None

    def __init__(
        self,
        flows: Optional[Union[ModuleType, List]] = None,
        blocks: Optional[Union[ModuleType, List]] = None,
        logger: Optional[Logger] = None,
        flows_to_ignore: List[str] = list(),
        blocks_to_ignore: List[str] = list(),
        default_flow: Optional[str] = None,
        default_block: Optional[str] = None,
    ):
        """
        init Orchestrator
        :param flows: data flows to init for current orchestrator
        :type flows: Optional[Union[ModuleType, List]]
        :param blocks: data blocks to init for current orchestrator
        :type blocks: Optional[Union[ModuleType, List]]
        :param logger: logger orchestrator
        :param flows_to_ignore: name of the flows to
         ignore which will not be initialized
        :type flows_to_ignore: List[str]
        :param blocks_to_ignore: name of the blocks to
         ignore which will not be initialized
        :type blocks_to_ignore: List[str]
        :param str default_flow: name of the flow that will be called
         if a non-existing flow is specified
        :param str default_block: name of the block that will be called
         if a non-existing block is specified
        """
        self.logger = logger or DEFAULT_LOGGER

        if flows:
            self._flows = self._generate_data(  # type: ignore
                data_to_process=flows,
                type_to_compare=(SyncFlow, AsyncFlow),
                attribute_to_get="name_flow",
                names_to_ignore=flows_to_ignore,
                type_data="flow",
            )
        elif self.flows:
            self._flows = self._generate_data(  # type: ignore
                data_to_process=self.flows,
                type_to_compare=(SyncFlow, AsyncFlow),
                attribute_to_get="name_flow",
                names_to_ignore=flows_to_ignore,
                type_data="flow",
            )
        if blocks:
            self._targets = self._generate_data(  # type: ignore
                data_to_process=blocks,
                type_to_compare=(SyncBlock, AsyncBlock),
                attribute_to_get="name_block",
                names_to_ignore=blocks_to_ignore,
                type_data="block",
            )
        elif self.blocks:
            self._targets = self._generate_data(  # type: ignore
                data_to_process=self.blocks,
                type_to_compare=(SyncBlock, AsyncBlock),
                attribute_to_get="name_block",
                names_to_ignore=blocks_to_ignore,
                type_data="block",
            )
        if not self._targets and not self._flows:
            raise NoDateException(
                "orchestrator",
                "There is no data for the orchestrator to work correctly",
            )
        self._validate_data()
        if default_flow:
            if not self._flows.get(default_flow):
                raise NotFoundDefaultError(
                    default_flow, list(self._flows.keys()), is_target=False
                )
            else:
                self._default_flow = default_flow
        if default_block:
            if not self._targets.get(default_block):
                raise NotFoundDefaultError(
                    default_block, list(self._targets.keys()), is_target=True
                )
            else:
                self._default_block = default_block
        self.logger.info(
            "Allowed flows to work: %s \n" "Allowed blocks to work: %s",
            self.get_list_flows(),
            self.get_list_blocks(),
        )

    def get_list_flows(self) -> List[str]:
        """
        :return: list allowed flows
        """
        return list(self._flows.keys())

    def get_list_blocks(self) -> List[str]:
        """
        :return:  list allowed blocks
        """
        return list(self._targets.keys())

    def _validate_data(self) -> None:
        """
        Checks the types of objects in the flow and targets the
        synchronous orchestrator not to redistribute a
        synchronous flows or blocks and vice versa
        :raise WorkTypeMismatchException: if incorrect type
        """

        def get_name(obj) -> str:
            if isinstance(obj, type):
                return obj.__name__
            return obj.__class__.__name__

        def check_type_dict_obj(
            dict_objects: Dict[str, Any], type_to_check: Type, is_target: bool
        ):
            for obj in dict_objects.values():
                if isinstance(obj, type):
                    if not issubclass(obj, type_to_check):
                        raise WorkTypeMismatchException(
                            base_class=self.__class__.__name__,
                            obj_class=get_name(obj),
                            is_target=is_target,
                        )
                elif not isinstance(obj, type_to_check):
                    raise WorkTypeMismatchException(
                        base_class=self.__class__.__name__,
                        obj_class=get_name(obj),
                        is_target=is_target,
                    )
                else:
                    continue

        check_type_dict_obj(self._targets, self._base_class_for_target, is_target=True)
        check_type_dict_obj(self._flows, self._base_class_for_flow, is_target=False)

    @staticmethod
    def _generate_data(
        data_to_process: Union[ModuleType, List],
        type_to_compare: Union[Type, Tuple[Type, ...]],
        attribute_to_get: str,
        names_to_ignore: Optional[List[str]] = list(),
        type_data: str = "block",
    ) -> Dict[
        str, Type[Union[SyncFlow, AsyncFlow, SyncBlock, AsyncBlock]]
    ]:  # pragma: no cover
        """
        The method prepares the data for the orchestrator to work
        Converts data from a list or retrieves from a module
        :param data_to_process:
        :param type_to_compare:
        :param attribute_to_get: name in object
        :param names_to_ignore: class names to ignore when processing
        :type names_to_ignore: List[str]
        :param type_data: name for information exception if will be raised
        :type type_data: str
        :raise TypeError: if the data to be processed contains
         the wrong type (and was not added to the exception)
        :raise UniqueNameException: if the data contains objects
         with non-unique identifiers
        :raise WrongTypeException: if a module is passed and
         not a list for generating data
        :raise NoDateException: if there is no data after processing
        :return: converted data to dictionary format to work with
        :rtype: Dict[str, Type[Union[Flow, SyncBlock, AsyncBlock]]]
        """
        _data = (
            dict()
        )  # type: Dict[str, Type[Union[SyncFlow, AsyncFlow, SyncBlock, AsyncBlock]]]
        if inspect.ismodule(data_to_process):
            for class_name, clazz in inspect.getmembers(
                data_to_process, inspect.isclass
            ):
                if class_name in names_to_ignore:
                    continue
                if not issubclass(clazz.__base__, type_to_compare):
                    raise TypeError(f"{type_data} is not inheritor {type_to_compare}")
                unique_identifier = getattr(clazz, attribute_to_get)
                if unique_identifier in names_to_ignore:
                    continue
                if _data.get(unique_identifier):
                    raise UniqueNameException(unique_identifier, type_data)
                _data[unique_identifier] = clazz
        elif isinstance(data_to_process, list):
            for obj in data_to_process:
                if isinstance(obj, type):
                    if obj.__name__ in names_to_ignore:
                        continue
                    if issubclass(obj.__base__, type_to_compare):
                        unique_identifier = getattr(obj, attribute_to_get)
                        if unique_identifier in names_to_ignore:
                            continue
                        if _data.get(unique_identifier):
                            raise UniqueNameException(unique_identifier, type_data)
                        _data[unique_identifier] = obj
                    else:
                        raise TypeError(
                            f"{type_data}: `{obj}` is not inheritor {type_to_compare} "
                        )
                elif isinstance(obj, type_to_compare):
                    if obj.__class__.__name__ in names_to_ignore:
                        continue
                    unique_identifier = getattr(obj, attribute_to_get)
                    if unique_identifier in names_to_ignore:
                        continue
                    if _data.get(unique_identifier):
                        raise UniqueNameException(unique_identifier, type_data)
                    _data[unique_identifier] = obj
                else:
                    raise TypeError(
                        f"{type_data}: `{obj}` is not inheritor {type_to_compare} "
                    )

        else:
            raise WrongTypeException(type_data, str(type(data_to_process)))
        if not _data:
            raise NoDateException(type_data)
        return _data

    def handle(
        self, message: BaseOrchServMsg, is_force_return: bool = False
    ) -> Optional[BaseOrchServMsg]:
        """
        Message processing method

        !!! the presence of a `target` in the message has higher priority
         than the presence of a `flow`
        :param message: message to process
        :type message: BaseOrchServMsg
        :param is_force_return: always return a message
        :type is_force_return: bool

        :return: message if an error occurred while processing the message
         if is_force_return is False
        :rtype: Optional[BaseOrchServMsg]
        """
        is_return_message = is_force_return
        if not isinstance(message, BaseOrchServMsg):
            raise TypeError(
                "Incorrect type `message`."
                " The `message` the message must be "
                f"of type `BaseOrchServMsg` and not {type(message)}"
            )
        self.logger.debug("Orchestrator. Started processing msg: %s", str(message))

        if message.get_flow() or message.get_target():
            if message.get_target():
                name_target = message.get_target()
                target = self._targets.get(name_target)
                if not target and not self._default_block:
                    is_return_message = True
                    self.logger.warning(
                        "Orchestrator. No suitable target was found to process "
                        "this message with target: %s."
                        " Available targets: %s",
                        name_target,
                        list(self._targets.keys()),
                    )
                else:
                    if not target:
                        target = self._targets.get(self._default_block)
                    if isinstance(target, type):
                        target = target(logger=self.logger)
                        self._targets[target.name_block] = target
                    try:
                        target.process(message)
                    except Exception as exc:
                        is_return_message = True
                        self.logger.warning(
                            "Orchestrator. Error processing msg `%s` in target `%s`."
                            " Error: %s",
                            str(message),
                            name_target,
                            str(exc),
                            exc_info=True,
                        )
            else:
                name_flow = message.get_flow()
                flow = self._flows.get(name_flow)
                if not flow and not self._default_flow:
                    is_return_message = True
                    self.logger.warning(
                        "Orchestrator. No suitable flow was found to process "
                        "this message with flow: %s."
                        " Available flows: %s",
                        name_flow,
                        list(self._flows.keys()),
                    )
                else:
                    if not flow:
                        flow = self._flows.get(self._default_flow)
                    if isinstance(flow, type):
                        flow = flow(logger=self.logger)
                        self._flows[flow.name_flow] = flow
                    try:
                        flow.to_go_with_the_flow(message)
                    except Exception as exc:
                        is_return_message = True
                        self.logger.warning(
                            "Orchestrator. Error processing msg %s in flow %s."
                            " Error: %s",
                            str(message),
                            name_flow,
                            str(exc),
                            exc_info=True,
                        )
        else:
            is_return_message = True
            self.logger.warning(
                "Orchestrator. Msg %s. "
                "Does not contain information about the type of"
                " processing and it was not processed",
                str(message),
            )
        self.logger.debug("Orchestrator. Finished processing msg: %s", str(message))
        if is_return_message:
            return message
        return None



import enum
import inspect
import warnings
from typing import Callable, Dict, Tuple, Union, get_args, get_origin

from orch_serv.exc import DataConsistencyError, ExtraAttributeError


class ParameterKind(enum.IntEnum):
    POSITIONAL_ONLY = 0
    POSITIONAL_OR_KEYWORD = 1
    VAR_POSITIONAL = 2
    KEYWORD_ONLY = 3
    VAR_KEYWORD = 4


def format_signature_parameters(parameters) -> str:  # pragma: no cover
    """
    Method copied from inspect for private usage
    """
    result = []
    render_pos_only_separator = False
    render_kw_only_separator = True
    for param in parameters.values():
        formatted = str(param)

        kind = param.kind

        if kind == ParameterKind.POSITIONAL_ONLY.value:
            render_pos_only_separator = True
        elif render_pos_only_separator:
            # It's not a positional-only parameter, and the flag
            # is set to 'True' (there were pos-only params before.)
            result.append("/")
            render_pos_only_separator = False

        if kind == ParameterKind.VAR_POSITIONAL.value:
            # OK, we have an '*args'-like parameter, so we won't need
            # a '*' to separate keyword-only arguments
            render_kw_only_separator = False
        elif kind == ParameterKind.KEYWORD_ONLY.value and render_kw_only_separator:
            # We have a keyword-only parameter to render and we haven't
            # rendered an '*args'-like parameter before, so add a '*'
            # separator to the parameters list ("foo(arg1, *, arg2)" case)
            result.append("*")
            # This condition should be only triggered once, so
            # reset the flag
            render_kw_only_separator = False

        result.append(formatted)

    if render_pos_only_separator:
        # There were only positional-only parameters, hence the
        # flag was not reset to 'False'
        result.append("/")

    return ", ".join(result)


def parse_signature(obj: Callable) -> Tuple[str, str]:
    """
    parse_signature of obj
    :param obj:
    :return: (attributes, returned)
    """
    signature = inspect.signature(obj)
    return format_signature_parameters(signature.parameters), inspect.formatannotation(
        signature.return_annotation
    )


def get_returned_value(obj: Callable):
    signature = inspect.signature(obj)
    if get_origin(signature.return_annotation) == tuple:
        return get_args(signature.return_annotation)
    return signature.return_annotation


def get_attributes_obj(obj: Callable) -> str:
    signature = inspect.signature(obj)
    return " ,".join(list(signature.parameters.keys()))


def is_optional(field):
    return get_origin(field) is Union and type(None) in get_args(field)


def validate_data_step(obj: Callable, additional_args: Dict):
    """
    Validate provided data for execution step
    :param Callable obj: obj to execution
    :param additional_args: kwargs for execution obj
    :return: nothing
    :raises ExtraAttributeError: if in kwargs
    """
    signature_func = inspect.signature(obj)
    provided_args = set(additional_args.keys())
    existed_attributes = set(signature_func.parameters.keys())
    extra_attributes = provided_args - existed_attributes
    if extra_attributes and not is_exist_keyword_variable(obj):
        raise ExtraAttributeError(
            obj.__name__, list(additional_args), get_attributes_obj(obj)
        )


def is_exist_keyword_variable(obj: Callable):
    """
    Check is exist **keyword variable in `obj`
    :param Callable obj: obj to execution
    :return: is exist in function **kwargs
    """
    for param in inspect.signature(obj).parameters.values():
        if str(param.kind) == "VAR_KEYWORD":
            return True
    return False


def validate_data_consistency(
    obj: Callable, return_previous_obj, additional_args: Dict
):
    """
    Function for validate data consistency between steps
    :param Callable obj: obj to execution
    :param return_previous_obj: annotations returned from previous step
    :param additional_args: kwargs for execution obj
    :return: nothing
    :raises DataConsistencyError: if not consistent data
    """
    signature_func = inspect.signature(obj)
    data_to_check = [return_previous_obj]
    if isinstance(return_previous_obj, tuple):
        data_to_check = list(return_previous_obj)
    if is_optional(return_previous_obj):
        # check is optional value
        for val in get_args(return_previous_obj):
            if val is not None:
                data_to_check = [val]
                break
    check_warnings = []
    errors = []

    for i, (param, val) in enumerate(signature_func.parameters.items()):
        # inspect._empty
        if i < len(data_to_check):
            # check in returned data
            if val.annotation != inspect._empty:  # type: ignore
                # check if the return type matches the expected type
                if data_to_check[i] == val.annotation:
                    continue
                else:
                    errors.append(
                        "Expected and passed type do not match."
                        f" Expected `{val.annotation}`, passed `{data_to_check[i]}`"
                    )
            else:
                check_warnings.append(f"Not exist annotations for variable `{param}`")
        else:
            # check in args or check is default
            if param in additional_args:
                # exist and we don't check type
                continue
            elif val.default != inspect._empty:  # type: ignore
                # not existing in additional_args but existing default value
                continue
            elif str(val.kind) == "VAR_KEYWORD":
                # for **kwargs
                continue
            else:
                # not returned, not existed in kwargs, not existing default value
                errors.append(
                    f"No value for variable `{param}`."
                    f" Provide a default value or pass a value when "
                    f"initializing the step"
                )
    if check_warnings:
        warnings.warn(
            f"orch_serv.Stepper."
            f" Warnings in time check object `{obj.__name__}`. "
            f"Warning(s):\n {','.join(check_warnings)}"
        )
    if errors:
        raise DataConsistencyError(obj.__name__, errors)
